
(** Schema common functionalities depending only on Schema_types *)

open Encodings
open Schema_types

(** {2 Exceptions} *)

exception XSD_validation_error of string
exception XSI_validation_error of string

(** {2 XSD printer *)

val print_schema : Format.formatter -> schema -> unit
val print_type : Format.formatter -> type_definition -> unit
val print_attribute : Format.formatter -> attribute_declaration -> unit
val print_element : Format.formatter -> element_declaration -> unit
val print_attribute_group :
  Format.formatter -> attribute_group_definition -> unit
val print_model_group : Format.formatter -> model_group_definition -> unit
val print_simple_type : Format.formatter -> simple_type_definition -> unit
val print_complex_type : Format.formatter -> complex_type_definition -> unit

(** {2 Deconstruction functions} *)

val name_of_element_declaration : element_declaration -> Utf8.t
val name_of_type_definition : type_definition -> Utf8.t
val name_of_simple_type_definition : simple_type_definition -> Utf8.t
val name_of_complex_type_definition : complex_type_definition -> Utf8.t
val name_of_attribute_declaration : attribute_declaration -> Utf8.t
val name_of_attribute_use : attribute_use -> Utf8.t
val name_of_attribute_group_definition : attribute_group_definition -> Utf8.t
val name_of_model_group_definition : model_group_definition -> Utf8.t
val name_of_particle : particle -> Utf8.t

val string_of_component_kind : component_kind -> string

val variety_of_simple_type_definition : simple_type_definition -> variety
val facets_of_simple_type_definition : simple_type_definition -> facets
val simple_type_of_type : type_definition -> simple_type_definition
val complex_type_of_type : type_definition -> complex_type_definition
val content_type_of_type : type_definition -> content_type

val get_type: Utf8.t -> schema -> type_definition
val get_attribute: Utf8.t -> schema -> attribute_declaration
val get_element: Utf8.t -> schema -> element_declaration
val get_attribute_group: Utf8.t -> schema -> attribute_group_definition
val get_model_group: Utf8.t -> schema -> model_group_definition

val get_component: component_kind -> Utf8.t -> schema -> component

val iter_types: schema -> (type_definition -> unit) -> unit
val iter_attributes: schema -> (attribute_declaration -> unit) -> unit
val iter_elements: schema -> (element_declaration -> unit) -> unit
val iter_attribute_groups:
  schema -> (attribute_group_definition -> unit) -> unit
val iter_model_groups: schema -> (model_group_definition -> unit) -> unit

val first_of_particle: particle -> first
val first_of_model_group: model_group -> first
val is_in_first: Utf8.t -> first -> bool
val nullable: particle -> bool

(** {2 Facets} *)

val merge_facets: facets -> facets -> facets

  (** restrict base new_facets new_name
   *  Implements simple type derivition by restriction *)
val restrict: simple_type_definition -> facets -> Utf8.t option ->
  simple_type_definition

(** {2 Miscellaneous} *) 
  (** empty set of facets (with the only exception of "whiteSpace", which is set
  to <`Collapse, true>, the mandatory value for all non string derived simple
  types) *)

val no_facets: facets
val anySimpleType: simple_type_definition
val anyType: type_definition

  (** @return the integer interval corrisponding to boundary facets *)
val get_interval: facets -> Intervals.t

  (** perform white space normalization according to XML recommendation *)
val normalize_white_space: white_space_handling -> Utf8.t -> Utf8.t

(** {2 event interface on top of CDuce values} *)

val stream_of_value: Value.t -> event Stream.t
val string_of_event: event -> string

